# Copyright (c) 2017 Ian Stapleton Cordasco
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module containing the tests for the URIBuilder object."""
try:
    from urllib.parse import parse_qs
except ImportError:
    from urlparse import parse_qs

import pytest

from rfc3986 import builder, uri_reference


def test_builder_default():
    """Verify the default values."""
    uribuilder = builder.URIBuilder()
    assert uribuilder.scheme is None
    assert uribuilder.userinfo is None
    assert uribuilder.host is None
    assert uribuilder.port is None
    assert uribuilder.path is None
    assert uribuilder.query is None
    assert uribuilder.fragment is None


def test_from_uri_reference():
    uri = uri_reference("http://foo.bar:1234/baz")
    uribuilder = builder.URIBuilder().from_uri(uri)
    assert uribuilder.scheme == "http"
    assert uribuilder.userinfo is None
    assert uribuilder.host == "foo.bar"
    assert uribuilder.port == "1234"
    assert uribuilder.path == "/baz"
    assert uribuilder.query is None
    assert uribuilder.fragment is None


def test_from_uri_string():
    uribuilder = builder.URIBuilder().from_uri("https://bar.foo:4321/boom")
    assert uribuilder.scheme == "https"
    assert uribuilder.userinfo is None
    assert uribuilder.host == "bar.foo"
    assert uribuilder.port == "4321"
    assert uribuilder.path == "/boom"
    assert uribuilder.query is None
    assert uribuilder.fragment is None


def test_repr():
    """Verify our repr looks like our class."""
    uribuilder = builder.URIBuilder()
    assert repr(uribuilder).startswith("URIBuilder(scheme=None")


@pytest.mark.parametrize(
    "scheme",
    [
        "https",
        "hTTps",
        "Https",
        "HtTpS",
        "HTTPS",
    ],
)
def test_add_scheme(scheme):
    """Verify schemes are normalized when added."""
    uribuilder = builder.URIBuilder().add_scheme(scheme)
    assert uribuilder.scheme == "https"


@pytest.mark.parametrize(
    "username, password, userinfo",
    [
        ("user", "pass", "user:pass"),
        ("user", None, "user"),
        ("user@domain.com", "password", "user%40domain.com:password"),
        ("user", "pass:word", "user:pass%3Aword"),
    ],
)
def test_add_credentials(username, password, userinfo):
    """Verify we normalize usernames and passwords."""
    uribuilder = builder.URIBuilder().add_credentials(username, password)
    assert uribuilder.userinfo == userinfo


def test_add_credentials_requires_username():
    """Verify one needs a username to add credentials."""
    with pytest.raises(ValueError):
        builder.URIBuilder().add_credentials(None, None)


@pytest.mark.parametrize(
    ["hostname", "expected_hostname"],
    [
        ("google.com", "google.com"),
        ("GOOGLE.COM", "google.com"),
        ("gOOgLe.COM", "google.com"),
        ("goOgLE.com", "google.com"),
        ("[::ff%etH0]", "[::ff%25etH0]"),
        ("[::ff%25etH0]", "[::ff%25etH0]"),
        ("[::FF%etH0]", "[::ff%25etH0]"),
    ],
)
def test_add_host(hostname, expected_hostname):
    """Verify we normalize hostnames in add_host."""
    uribuilder = builder.URIBuilder().add_host(hostname)
    assert uribuilder.host == expected_hostname


@pytest.mark.parametrize(
    "port",
    [
        -100,
        "-100",
        -1,
        "-1",
        65536,
        "65536",
        1000000,
        "1000000",
        "",
        "abc",
        "0b10",
    ],
)
def test_add_invalid_port(port):
    """Verify we raise a ValueError for invalid ports."""
    with pytest.raises(ValueError):
        builder.URIBuilder().add_port(port)


@pytest.mark.parametrize(
    "port, expected",
    [
        (0, "0"),
        ("0", "0"),
        (1, "1"),
        ("1", "1"),
        (22, "22"),
        ("22", "22"),
        (80, "80"),
        ("80", "80"),
        (443, "443"),
        ("443", "443"),
        (65535, "65535"),
        ("65535", "65535"),
    ],
)
def test_add_port(port, expected):
    """Verify we normalize our port."""
    uribuilder = builder.URIBuilder().add_port(port)
    assert uribuilder.port == expected


@pytest.mark.parametrize(
    "path",
    [
        "sigmavirus24/rfc3986",
        "/sigmavirus24/rfc3986",
    ],
)
def test_add_path(path):
    """Verify we normalize our path value."""
    uribuilder = builder.URIBuilder().add_path(path)
    assert uribuilder.path == "/sigmavirus24/rfc3986"


@pytest.mark.parametrize(
    "query_items, expected",
    [
        ({"a": "b c"}, "a=b+c"),
        ({"a": "b+c"}, "a=b%2Bc"),
        ([("a", "b c")], "a=b+c"),
        ([("a", "b+c")], "a=b%2Bc"),
        ([("a", "b"), ("c", "d")], "a=b&c=d"),
        ([("a", "b"), ("username", "@d")], "a=b&username=%40d"),
        ([("percent", "%")], "percent=%25"),
    ],
)
def test_add_query_from(query_items, expected):
    """Verify the behaviour of add_query_from."""
    uribuilder = builder.URIBuilder().add_query_from(query_items)
    assert uribuilder.query == expected


def test_add_query():
    """Verify we do not modify the provided query string."""
    uribuilder = builder.URIBuilder().add_query("username=@foo")
    assert uribuilder.query == "username=@foo"


def test_add_fragment():
    """Verify our handling of fragments."""
    uribuilder = builder.URIBuilder().add_fragment("section-2.5.1")
    assert uribuilder.fragment == "section-2.5.1"


@pytest.mark.parametrize(
    "uri, extend_with, expected_path",
    [
        ("https://api.github.com", "/users", "/users"),
        ("https://api.github.com", "/users/", "/users/"),
        ("https://api.github.com", "users", "/users"),
        ("https://api.github.com", "users/", "/users/"),
        ("", "users/", "/users/"),
        ("", "users", "/users"),
        ("?foo=bar", "users", "/users"),
        (
            "https://api.github.com/users/",
            "sigmavirus24",
            "/users/sigmavirus24",
        ),
        (
            "https://api.github.com/users",
            "sigmavirus24",
            "/users/sigmavirus24",
        ),
        (
            "https://api.github.com/users",
            "/sigmavirus24",
            "/users/sigmavirus24",
        ),
    ],
)
def test_extend_path(uri, extend_with, expected_path):
    """Verify the behaviour of extend_path."""
    uribuilder = (
        builder.URIBuilder()
        .from_uri(uri_reference(uri))
        .extend_path(extend_with)
    )
    assert uribuilder.path == expected_path


@pytest.mark.parametrize(
    "uri, extend_with, expected_query",
    [
        (
            "https://github.com",
            [("a", "b c"), ("d", "e&f")],
            {"a": ["b c"], "d": ["e&f"]},
        ),
        (
            "https://github.com?a=0",
            [("a", "b c"), ("d", "e&f")],
            {"a": ["0", "b c"], "d": ["e&f"]},
        ),
        (
            "https://github.com?a=0&e=f",
            [("a", "b c"), ("d", "e&f")],
            {"a": ["0", "b c"], "e": ["f"], "d": ["e&f"]},
        ),
        (
            "https://github.com",
            {"a": "b c", "d": "e&f"},
            {"a": ["b c"], "d": ["e&f"]},
        ),
        (
            "https://github.com?a=0",
            {"a": "b c", "d": "e&f"},
            {"a": ["0", "b c"], "d": ["e&f"]},
        ),
        (
            "https://github.com?a=0&e=f",
            {"a": "b c", "d": "e&f"},
            {"a": ["0", "b c"], "e": ["f"], "d": ["e&f"]},
        ),
    ],
)
def test_extend_query_with(uri, extend_with, expected_query):
    """Verify the behaviour of extend_query_with."""
    uribuilder = (
        builder.URIBuilder()
        .from_uri(uri_reference(uri))
        .extend_query_with(extend_with)
    )
    assert parse_qs(uribuilder.query) == expected_query


def test_finalize():
    """Verify the whole thing."""
    uri = (
        builder.URIBuilder()
        .add_scheme("https")
        .add_credentials("sigmavirus24", "not-my-re@l-password")
        .add_host("github.com")
        .add_path("sigmavirus24/rfc3986")
        .finalize()
        .unsplit()
    )
    expected = (
        "https://sigmavirus24:not-my-re%40l-password@github.com/"
        "sigmavirus24/rfc3986"
    )
    assert expected == uri


def test_geturl():
    """Verify the short-cut to the URL."""
    uri = (
        builder.URIBuilder()
        .add_scheme("https")
        .add_credentials("sigmavirus24", "not-my-re@l-password")
        .add_host("github.com")
        .add_path("sigmavirus24/rfc3986")
        .geturl()
    )
    expected = (
        "https://sigmavirus24:not-my-re%40l-password@github.com/"
        "sigmavirus24/rfc3986"
    )
    assert expected == uri



import itertools
import sys

import pytest

SNOWMAN = b"\xe2\x98\x83"

valid_hosts = [
    "[21DA:00D3:0000:2F3B:02AA:00FF:FE28:9C5A]",
    "[::1]",
    "[::1%25lo]",  # With ZoneID
    "[FF02:0:0:0:0:0:0:2%25en01]",  # With ZoneID
    "[FF02:30:0:0:0:0:0:5%25en1]",  # With ZoneID
    "[FF02:30:0:0:0:0:0:5%25%26]",  # With ZoneID
    "[FF02:30:0:0:0:0:0:5%2525]",  # With ZoneID
    "[21DA:D3:0:2F3B:2AA:FF:FE28:9C5A]",
    "[FE80::2AA:FF:FE9A:4CA2]",
    "[FF02::2]",
    "[FFFF::]",
    "[FF02:3::5]",
    "[FF02:0:0:0:0:0:0:2]",
    "[FF02:30:0:0:0:0:0:5]",
    "127.0.0.1",
    "www.example.com",
    "localhost",
    "http-bin.org",
    "%2Fvar%2Frun%2Fsocket",
    "6g9m8V6",  # Issue #48
]

invalid_hosts = [
    "[FF02::3::5]",  # IPv6 can only have one ::
    "[FADF:01]",  # Not properly compacted (missing a :)
    "[FADF:01%en0]",  # Not properly compacted (missing a :), Invalid ZoneID
    "[FADF::01%]",  # Empty Zone ID
    "localhost:80:80:80",  # Too many ports
    "256.256.256.256",  # Invalid IPv4 Address
    SNOWMAN.decode("utf-8"),
]

equivalent_hostnames = [
    "example.com",
    "eXample.com",
    "example.COM",
    "EXAMPLE.com",
    "ExAMPLE.com",
    "eXample.COM",
    "example.COM",
    "EXAMPLE.COM",
    "ExAMPLE.COM",
]
equivalent_schemes = [
    "https",
    "HTTPS",
    "HttPs",
    "hTTpS",
    "HtTpS",
]
equivalent_schemes_and_hostnames = list(
    itertools.product(
        equivalent_schemes,
        equivalent_hostnames,
    )
)


@pytest.fixture(params=valid_hosts)
def basic_uri(request):
    return "http://%s" % request.param


@pytest.fixture(params=equivalent_schemes_and_hostnames)
def uri_to_normalize(request):
    return "%s://%s" % request.param


@pytest.fixture(params=valid_hosts)
def basic_uri_with_port(request):
    return "ftp://%s:21" % request.param


@pytest.fixture(params=valid_hosts)
def uri_with_port_and_userinfo(request):
    return "ssh://user:pass@%s:22" % request.param


@pytest.fixture(params=valid_hosts)
def uri_with_port_and_tricky_userinfo(request):
    return "ssh://{}@{}:22".format("user%20!=:pass", request.param)


@pytest.fixture(params=valid_hosts)
def basic_uri_with_path(request):
    return "http://%s/path/to/resource" % request.param


@pytest.fixture(params=valid_hosts)
def uri_with_path_and_query(request):
    return "http://%s/path/to/resource?key=value" % request.param


@pytest.fixture(params=valid_hosts)
def uri_with_everything(request):
    return "https://user:pass@%s:443/path/to/resource?key=value#fragment" % (
        request.param
    )


@pytest.fixture(params=valid_hosts)
def relative_uri(request):
    return "//%s" % request.param


@pytest.fixture
def absolute_path_uri():
    return "/path/to/file"


@pytest.fixture(params=invalid_hosts)
def invalid_uri(request):
    return "https://%s" % request.param


@pytest.fixture(params=valid_hosts)
def uri_path_with_percent(request):
    return "https://%s/%% " % request.param


@pytest.fixture(params=valid_hosts)
def uri_query_with_percent(request):
    return "https://%s?a=%%" % request.param


@pytest.fixture(params=valid_hosts)
def uri_fragment_with_percent(request):
    return "https://%s#perc%%ent" % request.param


@pytest.fixture(params=equivalent_schemes)
def scheme_only(request):
    return "%s:" % request.param


sys.path.insert(0, ".")



# Copyright (c) 2015 Ian Stapleton Cordasco
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

import rfc3986
from . import base
from rfc3986 import exceptions
from rfc3986 import parseresult as pr

INVALID_PORTS = ["443:80", "443:80:443", "abcdef", "port", "43port"]

SNOWMAN = b"\xe2\x98\x83"
SNOWMAN_IDNA_HOST = "http://xn--n3h.com"


@pytest.mark.parametrize("port", INVALID_PORTS)
def test_port_parsing(port):
    with pytest.raises(exceptions.InvalidPort):
        rfc3986.urlparse(f"https://httpbin.org:{port}/get")


@pytest.mark.parametrize(
    "parts, unsplit",
    [
        (("https", None, "httpbin.org"), "https://httpbin.org"),
        (("https", "user", "httpbin.org"), "https://user@httpbin.org"),
        (
            ("https", None, "httpbin.org", 443, "/get"),
            "https://httpbin.org:443/get",
        ),
        (("HTTPS", None, "HTTPBIN.ORG"), "https://httpbin.org"),
    ],
)
def test_from_parts(parts, unsplit):
    uri = pr.ParseResult.from_parts(*parts)
    assert uri.unsplit() == unsplit


@pytest.mark.parametrize(
    "parts, unsplit",
    [
        (("https", None, "httpbin.org"), b"https://httpbin.org"),
        (("https", "user", "httpbin.org"), b"https://user@httpbin.org"),
        (
            ("https", None, "httpbin.org", 443, "/get"),
            b"https://httpbin.org:443/get",
        ),
        (("HTTPS", None, "HTTPBIN.ORG"), b"https://httpbin.org"),
    ],
)
def test_bytes_from_parts(parts, unsplit):
    uri = pr.ParseResultBytes.from_parts(*parts)
    assert uri.unsplit() == unsplit


class TestParseResultParsesURIs(base.BaseTestParsesURIs):
    test_class = pr.ParseResult


class TestParseResultUnsplits(base.BaseTestUnsplits):
    test_class = pr.ParseResult


def test_normalizes_uris_when_using_from_string(uri_to_normalize):
    """Verify we always get the same thing out as we expect."""
    result = pr.ParseResult.from_string(uri_to_normalize, lazy_normalize=False)
    assert result.scheme == "https"
    assert result.host == "example.com"


class TestStdlibShims:
    def test_uri_with_everything(self, uri_with_everything):
        uri = pr.ParseResult.from_string(uri_with_everything)
        assert uri.host == uri.hostname
        assert uri.netloc == uri.authority
        assert uri.query == uri.params
        assert uri.geturl() == uri.unsplit()


def test_creates_a_copy_with_a_new_path(uri_with_everything):
    uri = pr.ParseResult.from_string(uri_with_everything)
    new_uri = uri.copy_with(path="/parse/result/tests/are/fun")
    assert new_uri.path == "/parse/result/tests/are/fun"


def test_creates_a_copy_with_a_new_port(basic_uri):
    uri = pr.ParseResult.from_string(basic_uri)
    new_uri = uri.copy_with(port=443)
    assert new_uri.port == 443


def test_parse_result_encodes_itself(uri_with_everything):
    uri = pr.ParseResult.from_string(uri_with_everything)
    uribytes = uri.encode()
    encoding = uri.encoding
    assert uri.scheme.encode(encoding) == uribytes.scheme
    assert uri.userinfo.encode(encoding) == uribytes.userinfo
    assert uri.host.encode(encoding) == uribytes.host
    assert uri.port == uribytes.port
    assert uri.path.encode(encoding) == uribytes.path
    assert uri.query.encode(encoding) == uribytes.query
    assert uri.fragment.encode(encoding) == uribytes.fragment


class TestParseResultBytes:
    def test_handles_uri_with_everything(self, uri_with_everything):
        uri = pr.ParseResultBytes.from_string(uri_with_everything)
        assert uri.scheme == b"https"
        assert uri.path == b"/path/to/resource"
        assert uri.query == b"key=value"
        assert uri.fragment == b"fragment"
        assert uri.userinfo == b"user:pass"
        assert uri.port == 443
        assert isinstance(uri.authority, bytes) is True

    def test_raises_invalid_authority_for_invalid_uris(self, invalid_uri):
        with pytest.raises(exceptions.InvalidAuthority):
            pr.ParseResultBytes.from_string(invalid_uri)

    @pytest.mark.parametrize("port", INVALID_PORTS)
    def test_raises_invalid_port_non_strict_parse(self, port):
        with pytest.raises(exceptions.InvalidPort):
            pr.ParseResultBytes.from_string(
                f"https://httpbin.org:{port}/get", strict=False
            )

    def test_copy_with_a_new_path(self, uri_with_everything):
        uri = pr.ParseResultBytes.from_string(uri_with_everything)
        new_uri = uri.copy_with(path=b"/parse/result/tests/are/fun")
        assert new_uri.path == b"/parse/result/tests/are/fun"

    def test_copy_with_a_new_unicode_path(self, uri_with_everything):
        uri = pr.ParseResultBytes.from_string(uri_with_everything)
        pathbytes = b"/parse/result/tests/are/fun" + SNOWMAN
        new_uri = uri.copy_with(path=pathbytes.decode("utf-8"))
        assert new_uri.path == (b"/parse/result/tests/are/fun" + SNOWMAN)

    def test_unsplit(self):
        uri = pr.ParseResultBytes.from_string(
            b"http://" + SNOWMAN + b".com/path", strict=False
        )
        idna_encoded = SNOWMAN_IDNA_HOST.encode("utf-8") + b"/path"
        assert uri.unsplit(use_idna=True) == idna_encoded

    def test_eager_normalization_from_string(self):
        uri = pr.ParseResultBytes.from_string(
            b"http://" + SNOWMAN + b".com/path",
            strict=False,
            lazy_normalize=False,
        )
        assert uri.unsplit() == b"http:/path"

    def test_eager_normalization_from_parts(self):
        uri = pr.ParseResultBytes.from_parts(
            scheme="http",
            host=SNOWMAN.decode("utf-8"),
            path="/path",
            lazy_normalize=False,
        )
        assert uri.unsplit() == b"http:/path"



import sys

import pytest

import rfc3986
from rfc3986.exceptions import InvalidAuthority

try:
    import idna
except ImportError:
    idna = None


requires_idna = pytest.mark.skipif(
    idna is None, reason="This test requires the 'idna' module"
)
iri_to_uri = pytest.mark.parametrize(
    ["iri", "uri"],
    [
        ("http://Bücher.de", "http://xn--bcher-kva.de"),
        ("http://faß.de", "http://xn--fa-hia.de"),
        (
            "http://βόλος.com/β/ό?λ#ος",
            "http://xn--nxasmm1c.com/%CE%B2/%CF%8C?%CE%BB#%CE%BF%CF%82",
        ),
        ("http://ශ්\u200dරී.com", "http://xn--10cl1a0b660p.com"),
        ("http://نامه\u200cای.com", "http://xn--mgba3gch31f060k.com"),
        ("http://Bü:ẞ@gOoGle.com", "http://B%C3%BC:%E1%BA%9E@gOoGle.com"),
        ("http://ẞ.com:443", "http://xn--zca.com:443"),
        ("http://ẞ.foo.com", "http://xn--zca.foo.com"),
        ("http://Bẞ.com", "http://xn--b-qfa.com"),
        (
            "http+unix://%2Ftmp%2FTEST.sock/get",
            "http+unix://%2Ftmp%2FTEST.sock/get",
        ),
    ],
)


@requires_idna
@iri_to_uri
def test_encode_iri(iri, uri):
    assert rfc3986.iri_reference(iri).encode().unsplit() == uri


@iri_to_uri
def test_iri_equality(iri, uri):
    assert rfc3986.iri_reference(iri) == iri


def test_iri_equality_special_cases():
    assert rfc3986.iri_reference("http://Bü:ẞ@βόλος.com/β/ό?λ#ος") == (
        "http",
        "Bü:ẞ@βόλος.com",
        "/%CE%B2/%CF%8C",
        "%CE%BB",
        "%CE%BF%CF%82",
    )

    with pytest.raises(TypeError):
        rfc3986.iri_reference("http://ẞ.com") == 1


@requires_idna
@pytest.mark.parametrize(
    "iri",
    [
        "http://♥.net",
        "http://\u0378.net",
        pytest.param(
            "http://㛼.com",
            marks=pytest.mark.skipif(
                sys.version_info < (3, 3) and sys.maxunicode <= 0xFFFF,
                reason="Python configured without UCS-4 support",
            ),
        ),
    ],
)
def test_encode_invalid_iri(iri):
    iri_ref = rfc3986.iri_reference(iri)
    with pytest.raises(InvalidAuthority):
        iri_ref.encode()



from rfc3986.misc import merge_paths
from rfc3986.uri import URIReference


def test_merge_paths_with_base_path_without_base_authority():
    """Demonstrate merging with a base URI without an authority."""
    base = URIReference(
        scheme=None,
        authority=None,
        path="/foo/bar/bogus",
        query=None,
        fragment=None,
    )
    expected = "/foo/bar/relative"
    assert merge_paths(base, "relative") == expected


def test_merge_paths_with_base_authority_and_path():
    """Demonstrate merging with a base URI with an authority and path."""
    base = URIReference(
        scheme=None,
        authority="authority",
        path="/foo/bar/bogus",
        query=None,
        fragment=None,
    )
    expected = "/foo/bar/relative"
    assert merge_paths(base, "relative") == expected


def test_merge_paths_without_base_authority_or_path():
    """Demonstrate merging with a base URI without an authority or path."""
    base = URIReference(
        scheme=None, authority=None, path=None, query=None, fragment=None
    )
    expected = "/relative"
    assert merge_paths(base, "relative") == expected


def test_merge_paths_with_base_authority_without_path():
    """Demonstrate merging with a base URI without an authority or path."""
    base = URIReference(
        scheme=None,
        authority="authority",
        path=None,
        query=None,
        fragment=None,
    )
    expected = "/relative"
    assert merge_paths(base, "relative") == expected



from rfc3986.api import is_valid_uri
from rfc3986.api import normalize_uri
from rfc3986.api import uri_reference
from rfc3986.api import URIReference


def test_uri_reference():
    assert isinstance(uri_reference("http://example.com"), URIReference)


def test_is_valid_uri():
    assert is_valid_uri("http://example.com") is True


def test_normalize_uri():
    assert normalize_uri("HTTP://EXAMPLE.COM") == "http://example.com"



import pytest

from rfc3986.normalizers import encode_component
from rfc3986.normalizers import normalize_host
from rfc3986.normalizers import normalize_percent_characters
from rfc3986.normalizers import normalize_scheme
from rfc3986.normalizers import remove_dot_segments
from rfc3986.uri import URIReference


def test_normalize_scheme():
    assert "http" == normalize_scheme("htTp")
    assert "http" == normalize_scheme("http")
    assert "http" == normalize_scheme("HTTP")


def test_normalize_percent_characters():
    expected = "%3Athis_should_be_lowercase%DF%AB%4C"
    assert expected == normalize_percent_characters(
        "%3athis_should_be_lowercase%DF%ab%4c"
    )
    assert expected == normalize_percent_characters(
        "%3Athis_should_be_lowercase%DF%AB%4C"
    )
    assert expected == normalize_percent_characters(
        "%3Athis_should_be_lowercase%DF%aB%4C"
    )


paths = [
    # (Input, expected output)
    ("/foo/bar/.", "/foo/bar/"),
    ("/foo/bar/", "/foo/bar/"),
    ("/foo/bar", "/foo/bar"),
    ("./foo/bar", "foo/bar"),
    ("/./foo/bar", "/foo/bar"),
    ("/foo%20bar/biz%2Abaz", "/foo%20bar/biz%2Abaz"),
    ("../foo/bar", "foo/bar"),
    ("/../foo/bar", "/foo/bar"),
    ("a/./b/../b/%63/%7Bfoo%7D", "a/b/%63/%7Bfoo%7D"),
    ("//a/./b/../b/%63/%7Bfoo%7D", "//a/b/%63/%7Bfoo%7D"),
    ("mid/content=5/../6", "mid/6"),
    ("/a/b/c/./../../g", "/a/g"),
]


@pytest.fixture(params=paths)
def path_fixture(request):
    return request.param


@pytest.fixture(params=paths)
def uris(request):
    to_norm, normalized = request.param
    return (
        URIReference(None, None, to_norm, None, None),
        URIReference(None, None, normalized, None, None),
    )


def test_remove_dot_segments(path_fixture):
    to_normalize, expected = path_fixture
    assert expected == remove_dot_segments(to_normalize)


def test_normalized_equality(uris):
    assert uris[0] == uris[1]


def test_hostname_normalization():
    assert URIReference(None, "EXAMPLE.COM", None, None, None) == URIReference(
        None, "example.com", None, None, None
    )


@pytest.mark.parametrize(
    ["authority", "expected_authority"],
    [
        ("user%2aName@EXAMPLE.COM", "user%2AName@example.com"),
        ("[::1%eth0]", "[::1%25eth0]"),
    ],
)
def test_authority_normalization(authority, expected_authority):
    uri = URIReference(None, authority, None, None, None).normalize()
    assert uri.authority == expected_authority


def test_fragment_normalization():
    uri = URIReference(None, "example.com", None, None, "fiz%DF").normalize()
    assert uri.fragment == "fiz%DF"


@pytest.mark.parametrize(
    ["component", "encoded_component"],
    [
        ("/%", "/%25"),
        ("/~", "/~"),
        ("/%a", "/%25a"),
        ("/%ag", "/%25ag"),
        ("/%af", "/%af"),
        ("/%20/%", "/%2520/%25"),
        ("/%20%25", "/%20%25"),
        ("/%21%22%23%ah%12%ff", "/%2521%2522%2523%25ah%2512%25ff"),
    ],
)
def test_detect_percent_encoded_component(component, encoded_component):
    assert encode_component(component, "utf-8") == encoded_component


@pytest.mark.parametrize(
    ["host", "normalized_host"],
    [
        ("LOCALHOST", "localhost"),
        ("[::1%eth0]", "[::1%25eth0]"),
        ("[::1%25]", "[::1%2525]"),
        ("[::1%%25]", "[::1%25%25]"),
        ("[::1%25%25]", "[::1%25%25]"),
        ("[::Af%Ff]", "[::af%25Ff]"),
        ("[::Af%%Ff]", "[::af%25%Ff]"),
        ("[::Af%25Ff]", "[::af%25Ff]"),
    ],
)
def test_normalize_host(host, normalized_host):
    assert normalize_host(host) == normalized_host



"""Tests for the validators module."""
import pytest

import rfc3986
from rfc3986 import exceptions
from rfc3986 import validators


def test_defaults():
    """Verify the default Validator settings."""
    validator = validators.Validator()

    assert validator.required_components == {
        c: False for c in validator.COMPONENT_NAMES
    }
    assert validator.allow_password is True
    assert validator.allowed_schemes == set()
    assert validator.allowed_hosts == set()
    assert validator.allowed_ports == set()


def test_allowing_schemes():
    """Verify the ability to select schemes to be allowed."""
    validator = validators.Validator().allow_schemes("http", "https")

    assert "http" in validator.allowed_schemes
    assert "https" in validator.allowed_schemes


def test_allowing_hosts():
    """Verify the ability to select hosts to be allowed."""
    validator = validators.Validator().allow_hosts(
        "pypi.python.org",
        "pypi.org",
    )

    assert "pypi.python.org" in validator.allowed_hosts
    assert "pypi.org" in validator.allowed_hosts


def test_allowing_ports():
    """Verify the ability select ports to be allowed."""
    validator = validators.Validator().allow_ports("80", "100")

    assert "80" in validator.allowed_ports
    assert "100" in validator.allowed_ports


def test_requiring_invalid_component():
    """Verify that we validate required component names."""
    with pytest.raises(ValueError):
        validators.Validator().require_presence_of("frob")


def test_checking_validity_of_component():
    """Verify that we validate components we're validating."""
    with pytest.raises(ValueError):
        validators.Validator().check_validity_of("frob")


def test_use_of_password():
    """Verify the behaviour of {forbid,allow}_use_of_password."""
    validator = validators.Validator()
    assert validator.allow_password is True

    validator.forbid_use_of_password()
    assert validator.allow_password is False

    validator.allow_use_of_password()
    assert validator.allow_password is True


@pytest.mark.parametrize(
    "uri",
    [
        rfc3986.uri_reference("https://user:password@github.com"),
        rfc3986.uri_reference("https://user:password@github.com/path"),
        rfc3986.uri_reference("https://user:password@github.com/path?query"),
        rfc3986.uri_reference(
            "https://user:password@github.com/path?query#frag"
        ),
        rfc3986.uri_reference("//user:password@github.com"),
    ],
)
def test_forbidden_passwords(uri):
    """Verify that passwords are disallowed."""
    validator = validators.Validator().forbid_use_of_password()
    with pytest.raises(exceptions.PasswordForbidden):
        validator.validate(uri)


@pytest.mark.parametrize(
    "uri",
    [
        rfc3986.uri_reference("https://user@github.com"),
        rfc3986.uri_reference("https://user@github.com/path"),
        rfc3986.uri_reference("https://user@github.com/path?query"),
        rfc3986.uri_reference("https://user@github.com/path?query#frag"),
        rfc3986.uri_reference("//user@github.com"),
        rfc3986.uri_reference("//github.com"),
        rfc3986.uri_reference("https://github.com"),
    ],
)
def test_passwordless_uris_pass_validation(uri):
    """Verify password-less URLs validate properly."""
    validator = validators.Validator().forbid_use_of_password()
    validator.validate(uri)


@pytest.mark.parametrize(
    "uri",
    [
        rfc3986.uri_reference("https://"),
        rfc3986.uri_reference("/path/to/resource"),
    ],
)
def test_missing_host_component(uri):
    """Verify that missing host components cause errors."""
    validators.Validator().validate(uri)

    validator = validators.Validator().require_presence_of("host")
    with pytest.raises(exceptions.MissingComponentError):
        validator.validate(uri)


@pytest.mark.parametrize(
    "uri",
    [
        rfc3986.uri_reference("https://"),
        rfc3986.uri_reference("//google.com"),
        rfc3986.uri_reference("//google.com?query=value"),
        rfc3986.uri_reference("//google.com#fragment"),
        rfc3986.uri_reference("https://google.com"),
        rfc3986.uri_reference("https://google.com#fragment"),
        rfc3986.uri_reference("https://google.com?query=value"),
    ],
)
def test_missing_path_component(uri):
    """Verify that missing path components cause errors."""
    validator = validators.Validator().require_presence_of("path")
    with pytest.raises(exceptions.MissingComponentError):
        validator.validate(uri)


@pytest.mark.parametrize(
    "uri",
    [
        rfc3986.uri_reference("//google.com"),
        rfc3986.uri_reference("//google.com?query=value"),
        rfc3986.uri_reference("//google.com#fragment"),
    ],
)
def test_multiple_missing_components(uri):
    """Verify that multiple missing components are caught."""
    validator = validators.Validator().require_presence_of("scheme", "path")
    with pytest.raises(exceptions.MissingComponentError) as captured_exc:
        validator.validate(uri)
    exception = captured_exc.value
    assert 2 == len(exception.args[-1])


@pytest.mark.parametrize(
    "uri",
    [
        rfc3986.uri_reference("smtp://"),
        rfc3986.uri_reference("telnet://"),
    ],
)
def test_ensure_uri_has_a_scheme(uri):
    """Verify validation with allowed schemes."""
    validator = validators.Validator().allow_schemes("https", "http")
    with pytest.raises(exceptions.UnpermittedComponentError):
        validator.validate(uri)


@pytest.mark.parametrize(
    "uri, failed_component",
    [
        (rfc3986.uri_reference("git://github.com"), "scheme"),
        (rfc3986.uri_reference("http://github.com"), "scheme"),
        (rfc3986.uri_reference("ssh://gitlab.com"), "host"),
        (rfc3986.uri_reference("https://gitlab.com"), "host"),
    ],
)
def test_allowed_hosts_and_schemes(uri, failed_component):
    """Verify each of these fails."""
    validator = (
        validators.Validator()
        .allow_schemes(
            "https",
            "ssh",
        )
        .allow_hosts(
            "github.com",
            "git.openstack.org",
        )
    )
    with pytest.raises(exceptions.UnpermittedComponentError) as caught_exc:
        validator.validate(uri)

    exc = caught_exc.value
    assert exc.component_name == failed_component


@pytest.mark.parametrize(
    "uri",
    [
        rfc3986.uri_reference("https://github.com/sigmavirus24"),
        rfc3986.uri_reference("ssh://github.com/sigmavirus24"),
        rfc3986.uri_reference("ssh://ssh@github.com:22/sigmavirus24"),
        rfc3986.uri_reference("https://github.com:443/sigmavirus24"),
        rfc3986.uri_reference("https://gitlab.com/sigmavirus24"),
        rfc3986.uri_reference("ssh://gitlab.com/sigmavirus24"),
        rfc3986.uri_reference("ssh://ssh@gitlab.com:22/sigmavirus24"),
        rfc3986.uri_reference("https://gitlab.com:443/sigmavirus24"),
        rfc3986.uri_reference("https://bitbucket.org/sigmavirus24"),
        rfc3986.uri_reference("ssh://bitbucket.org/sigmavirus24"),
        rfc3986.uri_reference("ssh://ssh@bitbucket.org:22/sigmavirus24"),
        rfc3986.uri_reference("https://bitbucket.org:443/sigmavirus24"),
        rfc3986.uri_reference("https://git.openstack.org/sigmavirus24"),
        rfc3986.uri_reference("ssh://git.openstack.org/sigmavirus24"),
        rfc3986.uri_reference("ssh://ssh@git.openstack.org:22/sigmavirus24"),
        rfc3986.uri_reference("https://git.openstack.org:443/sigmavirus24"),
        rfc3986.uri_reference(
            "ssh://ssh@git.openstack.org:22/sigmavirus24?foo=bar#fragment"
        ),
        rfc3986.uri_reference(
            "ssh://git.openstack.org:22/sigmavirus24?foo=bar#fragment"
        ),
        rfc3986.uri_reference("ssh://git.openstack.org:22/?foo=bar#fragment"),
        rfc3986.uri_reference(
            "ssh://git.openstack.org:22/sigmavirus24#fragment"
        ),
        rfc3986.uri_reference("ssh://git.openstack.org:22/#fragment"),
        rfc3986.uri_reference("ssh://git.openstack.org:22/"),
        rfc3986.uri_reference(
            "ssh://ssh@git.openstack.org:22/?foo=bar#fragment"
        ),
        rfc3986.uri_reference(
            "ssh://ssh@git.openstack.org:22/sigmavirus24#fragment"
        ),
        rfc3986.uri_reference("ssh://ssh@git.openstack.org:22/#fragment"),
        rfc3986.uri_reference("ssh://ssh@git.openstack.org:22/"),
    ],
)
def test_successful_complex_validation(uri):
    """Verify we do not raise ValidationErrors for good URIs."""
    validators.Validator().allow_schemes("https", "ssh",).allow_hosts(
        "github.com",
        "bitbucket.org",
        "gitlab.com",
        "git.openstack.org",
    ).allow_ports("22", "443",).require_presence_of(
        "scheme",
        "host",
        "path",
    ).check_validity_of(
        "scheme",
        "userinfo",
        "host",
        "port",
        "path",
        "query",
        "fragment",
    ).validate(
        uri
    )


def test_invalid_uri_generates_error(invalid_uri):
    """Verify we catch invalid URIs."""
    uri = rfc3986.uri_reference(invalid_uri)
    with pytest.raises(exceptions.InvalidComponentsError):
        validators.Validator().check_validity_of("host").validate(uri)


def test_invalid_uri_with_invalid_path(invalid_uri):
    """Verify we catch multiple invalid components."""
    uri = rfc3986.uri_reference(invalid_uri)
    uri = uri.copy_with(path="#foobar")
    with pytest.raises(exceptions.InvalidComponentsError):
        validators.Validator().check_validity_of(
            "host",
            "path",
        ).validate(uri)


def test_validating_rfc_4007_ipv6_zone_ids():
    """Verify that RFC 4007 IPv6 Zone IDs are invalid
    host/authority but after normalization are valid
    """
    uri = rfc3986.uri_reference("http://[::1%eth0]")
    with pytest.raises(exceptions.InvalidComponentsError):
        validators.Validator().check_validity_of("host").validate(uri)

    uri = uri.normalize()
    assert uri.host == "[::1%25eth0]"

    validators.Validator().check_validity_of("host").validate(uri)



import pytest

from rfc3986 import exceptions
from rfc3986 import parseresult
from rfc3986 import uri_reference
from rfc3986 import urlparse


SNOWMAN = b"\xe2\x98\x83"
SNOWMAN_PARAMS = b"http://example.com?utf8=" + SNOWMAN
SNOWMAN_HOST = b"http://" + SNOWMAN + b".com"
SNOWMAN_IDNA_HOST = "http://xn--n3h.com"


def test_unicode_uri():
    url_bytestring = SNOWMAN_PARAMS
    unicode_url = url_bytestring.decode("utf-8")
    uri = uri_reference(unicode_url)
    assert uri.is_valid() is True
    assert uri == "http://example.com?utf8=%E2%98%83"


def test_unicode_uri_passed_as_bytes():
    url_bytestring = SNOWMAN_PARAMS
    uri = uri_reference(url_bytestring)
    assert uri.is_valid() is True
    assert uri == "http://example.com?utf8=%E2%98%83"


def test_unicode_authority():
    url_bytestring = SNOWMAN_HOST
    unicode_url = url_bytestring.decode("utf-8")
    uri = uri_reference(unicode_url)
    assert uri.is_valid() is False
    assert uri == unicode_url


def test_urlparse_a_unicode_hostname():
    url_bytestring = SNOWMAN_HOST
    unicode_url = url_bytestring.decode("utf-8")
    parsed = urlparse(url_bytestring)
    assert parsed.host == unicode_url[7:]


def test_urlparse_a_unicode_hostname_with_auth():
    url = b"http://userinfo@" + SNOWMAN + b".com"
    parsed = urlparse(url)
    assert parsed.userinfo == "userinfo"


def test_urlparse_idna_encoding_with_geturl():
    """https://github.com/python-hyper/rfc3986/issues/57"""
    parsed = urlparse("https://i❤.ws")
    encoded = parsed.encode("idna")
    assert encoded.encoding == "idna"
    assert encoded.geturl() == b"https://xn--i-7iq.ws"


def test_urlparse_an_invalid_authority_parses_port():
    url = "http://foo:b@r@[::1]:80/get"
    parsed = urlparse(url)
    assert parsed.port == 80
    assert parsed.userinfo == "foo:b@r"
    assert parsed.hostname == "[::1]"


def test_unsplit_idna_a_unicode_hostname():
    parsed = urlparse(SNOWMAN_HOST)
    assert parsed.unsplit(use_idna=True) == SNOWMAN_IDNA_HOST


def test_strict_urlparsing():
    with pytest.raises(exceptions.InvalidAuthority):
        parseresult.ParseResult.from_string(SNOWMAN_HOST)
